import numpy as np
import skia
from dataclasses import dataclass

from p5 import p5
from p5.core.attribs import stroke, fill
from p5.core.color import Color
from p5.core import constants
from p5.core.primitives import point, line
from p5.pmath.utils import *


@dataclass
class Style2D:
    background_color = (0.8, 0.8, 0.8, 1.0)
    fill_enabled = True
    stroke_enabled = True
    fill_color = (1, 1, 1, 1)
    stroke_color = (0, 0, 0)
    stroke_weight = 1

    tint_color = (0, 0, 0)
    tint_enabled = False
    ellipse_mode = constants.CENTER
    rect_mode = constants.CORNER
    color_parse_mode = "RGB"
    color_range = (255, 255, 255, 255)

    stroke_cap = skia.Paint.kRound_Cap
    stroke_join = skia.Paint.kMiter_Join

    def set_stroke_cap(self, c):
        if c == constants.ROUND:
            self.stroke_cap = skia.Paint.kRound_Cap
        elif c == constants.SQUARE:
            self.stroke_cap = skia.Paint.kButt_Cap
        elif c == constants.PROJECT:
            self.stroke_cap = skia.Paint.kSquare_Cap

    def set_stroke_join(self, j):
        if j == constants.ROUND:
            self.stroke_join = skia.Paint.kRound_Join
        elif j == constants.MITER:
            self.stroke_join = skia.Paint.kMiter_Join
        elif j == constants.BEVEL:
            self.stroke_join = skia.Paint.kBevel_Join


class SkiaRenderer:
    def __init__(self):
        self.canvas = None
        self.paint = None
        self.style = Style2D()
        self.style_stack = []
        self.path = None
        self.font = skia.Font()
        self.typeface = skia.Typeface.MakeDefault()
        self.font.setTypeface(self.typeface)
        # TODO: Implement curve tightness from curves.py
        #  use curve_tightness from the renderer instance or style object instead of global variable
        self.curve_tightness = 0

    # Transforms functions
    def push_matrix(self):
        self.canvas.save()

    def pop_matrix(self):
        self.canvas.restore()

    def rotate(self, theta):
        # angles are always in radians in p5py
        # TODO: add angle mode
        self.canvas.rotate(theta * 180 / PI)

    def translate(self, x, y, z):
        self.canvas.translate(x, y)

    def scale(self, x, y=None, z=None):
        if not y:
            y = x
        self.canvas.scale(x, y)

    def shear_x(self, theta):
        self.apply_matrix(np.array([1, tan(theta), 0, 0, 1, 0, 0, 0, 1]).reshape(3, 3))

    def shear_y(self, theta):
        self.apply_matrix(np.array([1, 0, 0, tan(theta), 1, 0, 0, 0, 1]).reshape(3, 3))

    def apply_matrix(self, transform_matrix):
        if transform_matrix.shape != (3, 3):
            raise ValueError(
                "Expected a (3,3) shaped matrix instead received",
                transform_matrix.shape,
            )
        curr = np.array([*self.canvas.getTotalMatrix()]).reshape(3, 3)
        self.canvas.setMatrix(skia.Matrix(np.matmul(curr, transform_matrix)))

    def reset_matrix(self):
        self.canvas.resetMatrix()

    # Rendering functions
    def _acute_arc_to_bezier(self, start, size):
        alpha = size / 2
        cos_alpha = cos(alpha)
        sin_alpha = sin(alpha)
        cot_alpha = 1 / tan(alpha)

        phi = start + alpha
        cos_phi = cos(phi)
        sin_phi = sin(phi)
        _lambda = (4 - cos_alpha) / 3
        mu = sin_alpha + (cos_alpha - _lambda) * cot_alpha

        return {
            "ax": round(cos(start), 7),
            "ay": round(sin(start), 7),
            "bx": round((_lambda * cos_phi + mu * sin_phi), 7),
            "by": round((_lambda * sin_phi - mu * cos_phi), 7),
            "cx": round((_lambda * cos_phi - mu * sin_phi), 7),
            "cy": round((_lambda * sin_phi + mu * cos_phi), 7),
            "dx": round(cos(start + size), 7),
            "dy": round(sin(start + size), 7),
        }

    def initialize_renderer(self, canvas, paint, path):
        self.canvas = canvas
        self.paint = paint
        self.path = path

        self.clear()

    def render_text(self, text, x, y):
        # full path works relative does not
        # assert (typeface is not None), "should not be NULL"
        # handle alignment manually
        if self.style.fill_enabled:
            self.paint.setStyle(skia.Paint.kFill_Style)
            self.paint.setColor(self.style.fill_color)
            self.canvas.drawPath(self.path, self.paint)

        if self.style.stroke_enabled:
            self.paint.setStyle(skia.Paint.kStroke_Style)
            self.paint.setColor(self.style.stroke_color)
            self.paint.setStrokeWidth(self.style.stroke_weight)
            self.canvas.drawPath(self.path, self.paint)
        self.canvas.drawSimpleText(text, x, y, self.font, self.paint)

    def load_font(self, path):
        """
        path: string
        Absolute path of the font file
        returns: skia.Typeface
        """
        typeface = skia.Typeface.MakeFromFile(path)
        return typeface

    def set_font(self, font):
        """
        Sets the current font to be used for rendering text
        """
        self.typeface = font
        self.font.setTypeface(self.typeface)

    def set_font_size(self, size):
        self.font.setSize(size)

    def render_image(self, img, *args):
        self.canvas.drawImage(img, *args)

    def clear(self):
        # Our screen will go black or all the pixels would be transparent
        self.canvas.clear(skia.Color(0, 0, 0, 0))

    def background(self, *args, **kwargs):
        # TODO: Add if args == pImage check
        # TODO: Add blend mode logic later
        self.push_matrix()
        self.reset_matrix()

        curr_fill = self.style.fill_color
        curr_fill_enabled = self.style.fill_enabled
        curr_stroke_enabled = self.style.stroke_enabled

        self.style.fill_enabled = True
        self.style.stroke_enabled = False
        self.style.fill_color = Color(*args, **kwargs).normalized
        self.rect(0, 0, *p5.sketch.size)
        self.style.fill_color = curr_fill
        self.style.fill_enabled = curr_fill_enabled
        self.style.stroke_enabled = curr_stroke_enabled

        self.pop_matrix()

    def render(self, fill=True, stroke=True, rewind=True):
        """
        Draw the path on current canvas using paint
        """
        # TODO: Check, do we really have to setColor, and setStyle on each render call
        # TODO: Explore ways to do this, such that it works with pGraphics as well
        if self.style.fill_enabled and fill:
            self.paint.setStyle(skia.Paint.kFill_Style)
            self.paint.setColor(skia.Color4f(*self.style.fill_color))
            self.canvas.drawPath(self.path, self.paint)

        if self.style.stroke_enabled and stroke:
            self.paint.setStrokeCap(self.style.stroke_cap)
            self.paint.setStrokeJoin(self.style.stroke_join)
            self.paint.setStyle(skia.Paint.kStroke_Style)
            self.paint.setColor(skia.Color4f(*self.style.stroke_color))
            self.paint.setStrokeWidth(self.style.stroke_weight)
            self.canvas.drawPath(self.path, self.paint)

        if rewind:
            self.path.rewind()

    def reset(self):
        self.reset_matrix()
        self.font.setSize(15)

    def line(self, path):
        x1, y1, x2, y2 = path[0].x, path[0].y, path[1].x, path[1].y
        self.path.moveTo(x1, y1)
        self.path.lineTo(x2, y2)
        self.render()

    def arc(self, x, y, w, h, start_angle, stop_angle, mode):
        rx = w / 2
        ry = h / 2
        epsilon = 0.00001
        curves = []

        x += rx
        y += ry

        while stop_angle - start_angle >= epsilon:
            arc_to_draw = min(stop_angle - start_angle, HALF_PI)
            curves.append(self._acute_arc_to_bezier(start_angle, arc_to_draw))
            start_angle += arc_to_draw

        if self.style.fill_enabled:
            for index, curve in enumerate(curves, 0):
                if index == 0:
                    self.path.moveTo(x + curve["ax"] * rx, y + curve["ay"] * ry)
                self.path.cubicTo(
                    x + curve["bx"] * rx,
                    y + curve["by"] * ry,
                    x + curve["cx"] * rx,
                    y + curve["cy"] * ry,
                    x + curve["dx"] * rx,
                    y + curve["dy"] * ry,
                )
            if mode == constants.PIE or not mode:
                self.path.lineTo(x, y)

            self.path.close()
            self.render(fill=True, stroke=False)

        if self.style.stroke_enabled:
            for index, curve in enumerate(curves, 0):
                if index == 0:
                    self.path.moveTo(x + curve["ax"] * rx, y + curve["ay"] * ry)
                self.path.cubicTo(
                    x + curve["bx"] * rx,
                    y + curve["by"] * ry,
                    x + curve["cx"] * rx,
                    y + curve["cy"] * ry,
                    x + curve["dx"] * rx,
                    y + curve["dy"] * ry,
                )
            if mode == constants.PIE:
                self.path.lineTo(x, y)
                self.path.close()
            elif mode == constants.CHORD:
                self.path.close()
            self.render(fill=False, stroke=True)

    def ellipse(self, x, y, w, h):
        kappa = 0.5522847498
        ox = w / 2 * kappa
        oy = h / 2 * kappa
        xe = x + w
        ye = y + h
        xm = x + w / 2
        ym = y + h / 2

        self.path.moveTo(x, ym)
        self.path.cubicTo(x, ym - oy, xm - ox, y, xm, y)
        self.path.cubicTo(xm + ox, y, xe, ym - oy, xe, ym)
        self.path.cubicTo(xe, ym + oy, xm + ox, ye, xm, ye)
        self.path.cubicTo(xm - ox, ye, x, ym + oy, x, ym)

        self.render()

    def circle(self, x, y, d):
        self.path.addCircle(x, y, d / 2)
        self.render()

    def point(self, x, y):
        s = self.style.stroke_color
        f = self.style.fill_color
        fe = self.style.fill_enabled

        # configure the settings beforehand we render
        # We draw an arc, and fill it to simulate a point
        self.style.fill_color = s
        self.style.fill_enabled = True
        self.style.stroke_enabled = False

        # This will render the point
        self.circle(x, y, self.style.stroke_weight)

        self.style.fill_color = f
        self.style.fill_enabled = fe
        self.style.stroke_enabled = True

    def quad(self, x1, y1, x2, y2, x3, y3, x4, y4):
        self.path.moveTo(x1, y1)
        self.path.lineTo(x2, y2)
        self.path.lineTo(x3, y3)
        self.path.lineTo(x4, y4)
        self.path.close()
        self.render()

    def rect(self, *args):
        x, y, w, h = args[:4]
        args = args[4:]
        tl = args[0] if len(args) >= 1 else None
        tr = args[1] if len(args) >= 2 else None
        br = args[2] if len(args) >= 3 else None
        bl = args[3] if len(args) >= 4 else None

        if tl is None:
            self.path.moveTo(x, y)
            self.path.lineTo(x + w, y)
            self.path.lineTo(x + w, y + h)
            self.path.lineTo(x, y + h)
            self.path.close()
            self.render()
            return

        if tr is None:
            tr = tl
        if br is None:
            br = tr
        if bl is None:
            bl = br

        absW = abs(w)
        absH = abs(h)
        hw = absW / 2
        hh = absH / 2
        if absW < 2 * tl:
            tl = hw
        if absH < 2 * tl:
            tl = hh
        if absW < 2 * tr:
            tr = hw
        if absH < 2 * tr:
            tr = hh
        if absW < 2 * br:
            br = hw
        if absH < 2 * br:
            br = hh
        if absW < 2 * bl:
            bl = hw
        if absH < 2 * bl:
            bl = hh

        self.path.moveTo(x + tl, y)
        self.path.arcTo(x + w, y, x + w, y + h, tr)
        self.path.arcTo(x + w, y + h, x, y + h, br)
        self.path.arcTo(x, y + h, x, y, bl)
        self.path.arcTo(x, y, x + w, y, tl)
        self.path.close()

        self.render()

    def triangle(self, *args):
        x1, y1, x2, y2, x3, y3 = args
        self.path.moveTo(x1, y1)
        self.path.lineTo(x2, y2)
        self.path.lineTo(x3, y3)
        self.path.close()

        self.render()

    def _do_fill_stroke_close(self, close_shape):
        if close_shape:
            self.path.close()
        self.render()

    def end_shape(
        self, mode, vertices, is_curve, is_bezier, is_quadratic, is_contour, shape_kind
    ):

        close_shape = mode == constants.CLOSE
        # NOT APPENDING AGAIN
        num_verts = len(vertices)
        if is_curve and shape_kind is None:
            if num_verts > 3:
                s = 1 - self.curve_tightness
                self.path.moveTo(vertices[1][0], vertices[1][1])
                for i in range(1, num_verts - 2):
                    b = []
                    v = vertices[i]
                    # logic for curve tightness is borrowed from
                    # https://github.com/processing/p5.js/blob/9cd186349cdb55c5faf28befff9c0d4a390e02ed/src/core/p5.Renderer2D.js#L767
                    b.append([v[0], v[1]])
                    b.append(
                        [
                            v[0]
                            + (s * vertices[i + 1][0] - s * vertices[i - 1][0]) / 6,
                            v[1]
                            + (s * vertices[i + 1][1] - s * vertices[i - 1][1]) / 6,
                        ]
                    )
                    b.append(
                        [
                            vertices[i + 1][0]
                            + (s * vertices[i][0] - s * vertices[i + 2][0]) / 6,
                            vertices[i + 1][1]
                            + (s * vertices[i][1] - s * vertices[i + 2][1]) / 6,
                        ]
                    )
                    b.append([vertices[i + 1][0], vertices[i + 1][1]])
                    self.path.cubicTo(
                        b[1][0], b[1][1], b[2][0], b[2][1], b[3][0], b[3][1]
                    )
                if close_shape:
                    self.path.lineTo(vertices[i + 1][0], vertices[i + 1][1])
                self._do_fill_stroke_close(close_shape)
        elif is_bezier and shape_kind is None:
            for i in range(num_verts):
                if vertices[i][-1].get("is_vert", None):
                    if (
                        vertices[i][-1].get("move_to", None)
                        or self.path.countVerbs() == 0
                    ):
                        self.path.moveTo(vertices[i][0], vertices[i][1])
                    else:
                        self.path.lineTo(vertices[i][0], vertices[i][1])
                else:
                    self.path.cubicTo(
                        vertices[i][0],
                        vertices[i][1],
                        vertices[i][2],
                        vertices[i][3],
                        vertices[i][4],
                        vertices[i][5],
                    )
            self._do_fill_stroke_close(close_shape)

        elif is_quadratic and shape_kind is None:
            for i in range(num_verts):
                if vertices[i][-1].get("is_vert", None):
                    if (
                        vertices[i][-1].get("move_to", None)
                        or self.path.countVerbs() == 0
                    ):
                        self.path.moveTo(vertices[i][0], vertices[i][1])
                    else:
                        self.path.lineTo(vertices[i][0], vertices[i][1])
                else:
                    self.path.quadTo(
                        vertices[i][0], vertices[i][1], vertices[i][2], vertices[i][3]
                    )
            self._do_fill_stroke_close(close_shape)
        else:
            if shape_kind == constants.POINTS:
                for i in range(num_verts):
                    v = vertices[i]
                    if self.style.stroke_enabled:
                        stroke(*v[6])
                    point(v[0], v[1])
            elif shape_kind == constants.LINES:
                for i in range(0, num_verts - 1, 2):
                    v = vertices[i]
                    if self.style.stroke_enabled:
                        stroke(*vertices[i + 1][6])
                    line(v[0], v[1], vertices[i + 1][0], vertices[i + 1][1])
            elif shape_kind == constants.TRIANGLES:
                for i in range(0, num_verts - 2, 3):
                    v = vertices[i]
                    self.path.moveTo(v[0], v[1])
                    self.path.lineTo(vertices[i + 1][0], vertices[i + 1][1])
                    self.path.lineTo(vertices[i + 2][0], vertices[i + 2][1])
                    self.path.close()
                    if self.style.fill_enabled:
                        fill(*vertices[i + 2][5])
                        self.render(fill=True, stroke=False, rewind=False)
                    if self.style.stroke_enabled:
                        stroke(*vertices[i + 2][6])
                        self.render(fill=False, stroke=True, rewind=True)
            elif shape_kind == constants.TRIANGLE_STRIP:
                for i in range(num_verts - 1):
                    v = vertices[i]
                    self.path.moveTo(vertices[i + 1][0], vertices[i + 1][1])
                    self.path.lineTo(v[0], v[1])
                    if self.style.stroke_enabled:
                        stroke(*vertices[i + 1][6])
                    if self.style.fill_enabled:
                        fill(*vertices[i + 1][5])
                    if i + 2 < num_verts:
                        self.path.lineTo(vertices[i + 2][0], vertices[i + 2][1])
                        if self.style.stroke_enabled:
                            stroke(*vertices[i + 2][6])
                        if self.style.fill_enabled:
                            fill(*vertices[i + 2][5])
                    self._do_fill_stroke_close(close_shape)
            elif shape_kind == constants.TRIANGLE_FAN:
                if num_verts > 2:
                    for i in range(2, num_verts):
                        v = vertices[i]
                        self.path.moveTo(vertices[0][0], vertices[0][1])
                        self.path.lineTo(vertices[i - 1][0], vertices[i - 1][1])
                        self.path.lineTo(v[0], v[1])
                        self.path.lineTo(vertices[0][0], vertices[0][1])

                        if i < num_verts - 1:
                            if (
                                self.style.fill_enabled and v[5] != vertices[i + 1][5]
                            ) or (
                                self.style.stroke_enabled and v[6] != vertices[i + 1][6]
                            ):
                                if self.style.fill_enabled:
                                    fill(*v[5])
                                    self.render(fill=True, stroke=False, rewind=False)
                                    fill(*vertices[i + 1][5])
                                if self.style.stroke_enabled:
                                    stroke(*v[6])
                                    self.render(fill=False, stroke=True, rewind=True)
                                    stroke(*vertices[i + 1][6])
                                self.path.close()
                                self.path.rewind()
                    self._do_fill_stroke_close(close_shape)
            elif shape_kind == constants.QUADS:
                for i in range(0, num_verts - 3, 4):
                    v = vertices[i]
                    self.path.moveTo(v[0], v[1])
                    for j in range(1, 4):
                        self.path.lineTo(vertices[i + j][0], vertices[i + j][1])
                    self.path.lineTo(v[0], v[1])
                    if self.style.fill_enabled:
                        fill(*vertices[i + 3][5])
                    if self.style.stroke_enabled:
                        stroke(*vertices[i + 3][6])
                    self._do_fill_stroke_close(close_shape)
            elif shape_kind == constants.QUAD_STRIP:
                if num_verts > 3:
                    for i in range(0, num_verts - 1, 2):
                        v = vertices[i]
                        if i + 3 < num_verts:
                            self.path.moveTo(vertices[i + 2][0], vertices[i + 2][1])
                            self.path.lineTo(v[0], v[1])
                            self.path.lineTo(vertices[i + 1][0], vertices[i + 1][1])
                            self.path.lineTo(vertices[i + 3][0], vertices[i + 3][1])
                            if self.style.fill_enabled:
                                fill(*vertices[i + 3][5])
                            if self.style.stroke_enabled:
                                stroke(*vertices[i + 3][6])
                        else:
                            self.path.moveTo(v[0], v[1])
                            self.path.lineTo(vertices[i + 1][0], vertices[i + 1][1])
                        self._do_fill_stroke_close(close_shape)
            else:
                self.path.moveTo(vertices[0][0], vertices[0][1])
                for i in range(1, num_verts):
                    v = vertices[i]
                    if v[-1].get("is_vert", None):
                        if v[-1].get("move_to", None) or self.path.countVerbs() == 0:
                            self.path.moveTo(v[0], v[1])
                        else:
                            self.path.lineTo(v[0], v[1])
                self._do_fill_stroke_close(close_shape)

    # Fonts functions

    # Images functions



# This file adds compatibility for processing API

from .userspace import no_loop, save_frame


def noLoop():
    """Stop draw() from being continuously called.

    By default, the sketch continuously calls `draw()` as long as it
    runs. Calling `noLoop()` stops draw() from being called the next
    time. Note that this only prevents execution of the code inside
    `draw()` and the user can manipulate the screen contents through
    event handlers like `mousePressed()`, etc.

    """
    no_loop()


def saveFrame():
    """Save a numbered sequence of images whenever the function is run.

    Saves a numbered sequence of images, one image each time the
    function is run. To save an image that is identical to the display
    window, run the function at the end of :meth:`p5.draw` or within
    mouse and key events such as :meth:`p5.mouse_pressed` and
    :meth:`p5.key_pressed`.

    If saveFrame() is used without parameters, it will save files as
    screen-0000.png, screen-0001.png, and so on. Append a file
    extension, to indicate the file format to be used. Image files are
    saved to the sketch's folder. Alternatively, the files can be
    saved to any location on the computer by using an absolute path
    (something that starts with / on Unix and Linux, or a drive letter
    on Windows).

    :param filename: name (or name with path) of the image file.
        (defaults to ``screen.png``)

    :type filename: str

    """
    save_frame()



import unittest

from p5.pmath.curves import (
    bezier_point,
    bezier_tangent,
    curve_point,
    curve_tangent,
    quadratic_point,
)
from p5.pmath.vector import Point


class TestCurves(unittest.TestCase):
    def test_bezier_point(self):
        p1 = (30, 20)
        p2 = (80, 5)
        p3 = (80, 75)
        p4 = (30, 75)

        self.assertEqual(bezier_point(p1, p2, p3, p4, 0.5), (67.5, 41.875))

        self.assertEqual(bezier_point(p1, p2, p3, p4, 0.9), (43.5, 73.055))

        self.assertEqual(bezier_point(30, 80, 80, 30, 0.5), 67.5)

    def test_bezier_tangent(self):
        p1 = (30, 20)
        p2 = (80, 5)
        p3 = (80, 75)
        p4 = (30, 75)

        self.assertEqual(bezier_tangent(p1, p2, p3, p4, 0.5), (0.0, 93.75))

        self.assertEqual(bezier_tangent(p1[0], p2[0], p3[0], p4[0], 0.5), 0.0)

    def test_curve_point(self):
        p1 = (73, 24)
        p2 = (73, 61)
        p3 = (15, 65)
        p4 = (15, 65)
        self.assertEqual(curve_point(p1, p2, p3, p4, 0.5), (44.0, 65.3125))

    def test_curve_tangent(self):
        p1 = (73, 24)
        p2 = (73, 61)
        p3 = (15, 65)
        p4 = (15, 65)
        self.assertEqual(curve_tangent(p1, p2, p3, p4, 0.5), (-72.5, 0.375))

    def test_quadratic_point(self):
        p1 = (20, 20)
        p2 = (80, 20)
        p3 = (50, 50)
        self.assertEqual(quadratic_point(p1, p2, p3, 0.5), (57.5, 27.5))


if __name__ == "__main__":
    unittest.main()



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
import math
from p5.pmath import matrix
from .shaders2d import src_default, src_fbuffer

import builtins

from vispy import gloo
from vispy.gloo import Program
from vispy.gloo import Texture2D
from vispy.gloo import VertexBuffer

from contextlib import contextmanager
from .shaders2d import src_texture
from .shaders2d import src_line
from .openglrenderer import OpenGLRenderer, get_render_primitives, COLOR_WHITE
from .shape import PShape, Arc
from p5.core.constants import SType


class VispyRenderer2D(OpenGLRenderer):
    def __init__(self):
        super().__init__(src_fbuffer, src_default)
        self.texture_prog = Program(src_texture.vert, src_texture.frag)
        self.texture_prog["texcoord"] = self.fbuf_texcoords
        self.line_prog = None
        self.modelview_matrix = np.identity(4)

    def reset_view(self):
        self.viewport = (
            0,
            0,
            int(builtins.width * builtins.pixel_x_density),
            int(builtins.height * builtins.pixel_y_density),
        )
        self.texture_viewport = (
            0,
            0,
            builtins.width,
            builtins.height,
        )

        gloo.set_viewport(*self.viewport)  # pylint: disable=no-member

        cz = (builtins.height / 2) / math.tan(math.radians(30))
        self.projection_matrix = matrix.perspective_matrix(
            math.radians(60), builtins.width / builtins.height, 0.1 * cz, 10 * cz
        )
        self.modelview_matrix = matrix.translation_matrix(
            -builtins.width / 2, builtins.height / 2, -cz
        )
        self.modelview_matrix = self.modelview_matrix.dot(
            matrix.scale_transform(1, -1, 1)
        )

        self.transform_matrix = np.identity(4)

        self.default_prog["modelview"] = self.modelview_matrix.T.flatten()
        self.default_prog["projection"] = self.projection_matrix.T.flatten()

        self.texture_prog["modelview"] = self.modelview_matrix.T.flatten()
        self.texture_prog["projection"] = self.projection_matrix.T.flatten()

        self.line_prog = Program(src_line.vert, src_line.frag)

        self.line_prog["modelview"] = self.modelview_matrix.T.flatten()
        self.line_prog["projection"] = self.projection_matrix.T.flatten()
        self.line_prog["height"] = builtins.height

        self.fbuffer_tex_front = Texture2D((builtins.height, builtins.width, 3))
        self.fbuffer_tex_back = Texture2D((builtins.height, builtins.width, 3))

        for buf in [self.fbuffer_tex_front, self.fbuffer_tex_back]:
            self.fbuffer.color_buffer = buf
            with self.fbuffer:
                self.clear()

    def clear(self, color=True, depth=True):
        """Clear the renderer background."""
        gloo.set_state(  # pylint: disable=no-member
            clear_color=self.style.background_color
        )
        gloo.clear(color=color, depth=depth)  # pylint: disable=no-member

    def _comm_toggles(self, state=True):
        gloo.set_state(blend=state)  # pylint: disable=no-member
        gloo.set_state(depth_test=state)  # pylint: disable=no-member

        if state:
            gloo.set_state(  # pylint: disable=no-member
                blend_func=("src_alpha", "one_minus_src_alpha")
            )
            gloo.set_state(depth_func="lequal")  # pylint: disable=no-member

    @contextmanager
    def draw_loop(self):
        """The main draw loop context manager."""

        self.transform_matrix = np.identity(4)

        self.default_prog["modelview"] = self.modelview_matrix.T.flatten()
        self.default_prog["projection"] = self.projection_matrix.T.flatten()

        self.fbuffer.color_buffer = self.fbuffer_tex_back

        with self.fbuffer:
            gloo.set_viewport(*self.texture_viewport)  # pylint: disable=no-member
            self._comm_toggles()
            self.fbuffer_prog["texture"] = self.fbuffer_tex_front
            self.fbuffer_prog.draw("triangle_strip")

            yield

            self.flush_geometry()
            self.transform_matrix = np.identity(4)

        gloo.set_viewport(*self.viewport)  # pylint: disable=no-member
        self._comm_toggles(False)
        self.clear()
        self.fbuffer_prog["texture"] = self.fbuffer_tex_back
        self.fbuffer_prog.draw("triangle_strip")

        self.fbuffer_tex_front, self.fbuffer_tex_back = (
            self.fbuffer_tex_back,
            self.fbuffer_tex_front,
        )

    def _add_to_draw_queue(
        self, stype, vertices, idx, fill, stroke, stroke_weight, stroke_cap, stroke_join
    ):
        """Adds shape of stype to draw queue"""
        if stype == "lines":
            self.draw_queue.append(
                (stype, (vertices, idx, stroke, stroke_weight, stroke_cap, stroke_join))
            )
        else:
            self.draw_queue.append((stype, (vertices, idx, fill)))

    def render(self, shape):
        fill = shape.fill.normalized if shape.fill else None
        stroke = shape.stroke.normalized if shape.stroke else None
        stroke_weight = shape.stroke_weight
        stroke_cap = shape.stroke_cap
        stroke_join = shape.stroke_join

        obj_list = get_render_primitives(shape)
        for obj in obj_list:
            stype, vertices, idx = obj
            # Convert 2D vertices to 3D by adding "0" column, needed for further transformations
            if len(vertices[0]) == 2:
                vertices = np.hstack([vertices, np.zeros((len(vertices), 1))])
            # Transform vertices
            vertices = self._transform_vertices(
                np.hstack([vertices, np.ones((len(vertices), 1))]),
                shape._matrix,
                self.transform_matrix,
            )
            # Add to draw queue
            self._add_to_draw_queue(
                stype,
                vertices,
                idx,
                fill,
                stroke,
                stroke_weight,
                stroke_cap,
                stroke_join,
            )

    def flush_geometry(self):
        """Flush all the shape geometry from the draw queue to the GPU."""
        current_queue = []
        for index, shape in enumerate(self.draw_queue):
            current_shape = self.draw_queue[index][0]
            current_queue.append(self.draw_queue[index][1])

            if current_shape == "lines":
                self.render_line(current_queue)
            else:
                self.render_default(current_shape, current_queue)

            current_queue = []

        self.draw_queue = []

    def render_line(self, queue):
        """
        This rendering algorithm works by tesselating the line into
        multiple triangles.

        Reference: https://blog.mapbox.com/drawing-antialiased-lines-with-opengl-8766f34192dc
        """

        if len(queue) == 0:
            return

        pos = []
        posPrev = []
        posCurr = []
        posNext = []
        markers = []
        side = []

        linewidth = []
        join_type = []
        cap_type = []
        color = []

        stroke_cap_codes = {"PROJECT": 0, "SQUARE": 1, "ROUND": 2}

        stroke_join_codes = {"MITER": 0, "BEVEL": 1, "ROUND": 2}

        for line in queue:
            if len(line[1]) == 0:
                continue

            for segment in line[1]:
                for i in range(
                    len(segment) - 1
                ):  # the data is sent to renderer in line segments
                    for j in [0, 0, 1, 0, 1, 1]:  # all the vertices of triangles
                        if i + j - 1 >= 0:
                            posPrev.append(line[0][segment[i + j - 1]])
                        else:
                            posPrev.append(line[0][segment[i + j]])

                        if i + j + 1 < len(segment):
                            posNext.append(line[0][segment[i + j + 1]])
                        else:
                            posNext.append(line[0][segment[i + j]])

                        posCurr.append(line[0][segment[i + j]])

                    # Is the vertex up/below the line segment
                    markers.extend([1.0, -1.0, -1.0, -1.0, 1.0, -1.0])
                    # Left or right side of the segment
                    side.extend([1.0, 1.0, -1.0, 1.0, -1.0, -1.0])
                    # Left vertex of each segment
                    pos.extend([line[0][segment[i]]] * 6)
                    linewidth.extend([line[3]] * 6)
                    join_type.extend([stroke_join_codes[line[5]]] * 6)
                    cap_type.extend([stroke_cap_codes[line[4]]] * 6)
                    color.extend([line[2]] * 6)

        if len(pos) == 0:
            return

        posPrev = np.array(posPrev, np.float32)
        posCurr = np.array(posCurr, np.float32)
        posNext = np.array(posNext, np.float32)
        markers = np.array(markers, np.float32)
        side = np.array(side, np.float32)
        pos = np.array(pos, np.float32)
        linewidth = np.array(linewidth, np.float32)
        join_type = np.array(join_type, np.float32)
        cap_type = np.array(cap_type, np.float32)
        color = np.array(color, np.float32)

        self.line_prog["pos"] = gloo.VertexBuffer(pos)
        self.line_prog["posPrev"] = gloo.VertexBuffer(posPrev)
        self.line_prog["posCurr"] = gloo.VertexBuffer(posCurr)
        self.line_prog["posNext"] = gloo.VertexBuffer(posNext)
        self.line_prog["marker"] = gloo.VertexBuffer(markers)
        self.line_prog["side"] = gloo.VertexBuffer(side)
        self.line_prog["linewidth"] = gloo.VertexBuffer(linewidth)
        self.line_prog["join_type"] = gloo.VertexBuffer(join_type)
        self.line_prog["cap_type"] = gloo.VertexBuffer(cap_type)
        self.line_prog["color"] = gloo.VertexBuffer(color)

        self.line_prog.draw("triangles")

    def render_image(self, image, location, size):
        """Render the image.

        :param image: image to be rendered
        :type image: builtins.Image

        :param location: top-left corner of the image
        :type location: tuple | list | builtins.Vector

        :param size: target size of the image to draw.
        :type size: tuple | list | builtins.Vector
        """
        self.flush_geometry()

        self.texture_prog["fill_color"] = (
            self.style.tint_color if self.style.tint_enabled else COLOR_WHITE
        )
        self.texture_prog["transform"] = self.transform_matrix.T.flatten()

        x, y = location
        sx, sy = size
        imx, imy = image.size
        data = np.zeros(
            4, dtype=[("position", np.float32, 2), ("texcoord", np.float32, 2)]
        )
        data["texcoord"] = np.array(
            [[0.0, 1.0], [1.0, 1.0], [0.0, 0.0], [1.0, 0.0]], dtype=np.float32
        )
        data["position"] = np.array(
            [[x, y + sy], [x + sx, y + sy], [x, y], [x + sx, y]], dtype=np.float32
        )

        self.texture_prog["texture"] = image._texture
        self.texture_prog.bind(VertexBuffer(data))
        self.texture_prog.draw("triangle_strip")

    def cleanup(self):
        """Run the clean-up routine for the renderer.

        This method is called when all drawing has been completed and the
        program is about to exit.

        """
        OpenGLRenderer.cleanup(self)
        self.line_prog.delete()

    def render_shape(self, shape):
        self.render(shape)
        for child_shape in shape.children:
            self.render_shape(child_shape)

    def line(self, *args):
        path = args[0]
        self.render_shape(PShape(vertices=path, shape_type=SType.LINES))

    def bezier(self, *args):
        vertices = args[0]
        self.render_shape(PShape(vertices=vertices, shape_type=SType.LINE_STRIP))

    def curve(self, *args):
        vertices = args[0]
        self.render_shape(PShape(vertices=vertices, shape_type=SType.LINE_STRIP))

    def triangle(self, *args):
        path = args[0]
        self.render_shape(PShape(vertices=path, shape_type=SType.TRIANGLES))

    def quad(self, *args):
        path = args[0]
        self.render_shape(PShape(vertices=path, shape_type=SType.QUADS))

    def arc(self, *args):
        center = args[0]
        dim = args[1]
        start_angle = args[2]
        stop_angle = args[3]
        mode = args[4]

        self.render_shape(Arc(center, dim, start_angle, stop_angle, mode))

    def shape(self, vertices, contours, shape_type, *args):
        """Render a Pshape"""
        self.render_shape(
            PShape(vertices=vertices, contours=contours, shape_type=shape_type)
        )



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import xml.etree.ElementTree as etree
import math
import re
import numpy as np

from . import Color
from ..sketch.Vispy2DRenderer.shape import PShape
from . import primitives
from . import transforms
from ..pmath import matrix
from .constants import ROUND

default_values = {  # default values of SVG attributes
    "stroke-width": 1,
    "stroke-color": Color(0, 0, 0),
    "stroke-join": 0,
    "stroke-cap": 1,
    "stroke": "none",
    "fill": "none",
}


def get_style(element, style):
    # style parser for svg
    values = element.get("style")

    if element.get(style):
        return element.get(style)

    if values:
        for s in values.split(";"):
            if style in s:
                value = s.split(":")[1]

                if style == "stroke" or style == "fill":
                    return value
                if style == "stroke-width":
                    return int(value.replace("px", ""))
                if style == "stroke-opacity":
                    return float(value)

    if style in default_values.keys():
        return default_values[style]
    else:
        return None


def parse_rect(element):
    width = float(element.get("width"))
    height = float(element.get("height"))
    x = float(element.get("x"))
    y = float(element.get("y"))

    fill = Color(get_style(element, "fill"))
    stroke_weight = get_style(element, "stroke-width")
    stroke = Color(get_style(element, "stroke"))
    stroke_cap = get_style(element, "stroke-cap")

    return PShape(
        vertices=[(x, y), (x + width, y), (x + width, y + height), (x, y + height)],
        children=[],
        fill_color=fill,
        stroke_weight=stroke_weight,
        stroke_color=stroke,
        stroke_cap=stroke_cap,
        stroke_join=default_values["stroke-join"],
    )


def parse_circle(element):
    cx = float(element.get("cx"))
    cy = float(element.get("cy"))
    r = float(element.get("r"))

    fill = Color(get_style(element, "fill"))
    stroke_weight = get_style(element, "stroke-width")
    stroke = Color(get_style(element, "stroke"))
    stroke_cap = get_style(element, "stroke-cap")

    return primitives.Arc(
        (cx, cy),
        (r / 2, r / 2),
        0,
        2 * math.pi,
        "CHORD",
        fill_color=fill,
        stroke_weight=stroke_weight,
        stroke_color=stroke,
        stroke_cap=ROUND,
    )


def parse_line(element):
    x1 = float(element.get("x1"))
    y1 = float(element.get("y1"))
    x2 = float(element.get("x2"))
    y2 = float(element.get("y2"))

    fill = Color(get_style(element, "fill"))
    stroke_weight = get_style(element, "stroke-width")
    stroke = Color(get_style(element, "stroke"))
    stroke_cap = get_style(element, "stroke-cap")

    return PShape(
        vertices=[(x1, y1), (x2, y2)],
        fill_color=fill,
        stroke_weight=stroke_weight,
        stroke_color=stroke,
        stroke_cap=stroke_cap,
        stroke_join=default_values["stroke-join"],
    )


def parse_ellipse(element):
    cx = float(element.get("cx"))
    cy = float(element.get("cx"))
    rx = float(element.get("rx"))
    ry = float(element.get("ry"))

    fill = Color(get_style(element, "fill"))
    stroke_weight = get_style(element, "stroke-width")
    stroke = Color(get_style(element, "stroke"))
    stroke_cap = get_style(element, "stroke-cap")

    return primitives.Arc(
        (cx, cy),
        (rx / 2, ry / 2),
        0,
        2 * math.pi,
        "CHORD",
        fill_color=fill,
        stroke_weight=stroke_weight,
        stroke_color=stroke,
        stroke_cap=ROUND,
        stroke_join=default_values["stroke-join"],
    )


parser_function = {
    # tag: parser
    "rect": parse_rect,
    "circle": parse_circle,
    "line": parse_line,
    "ellipse": parse_ellipse,
    # "path": parse_path
}


def load_shape(filename):
    """
    Loads the given .svg file and converts it into
    PShape object.

    :param filename: link to .svg file
    :type filename: str

    """
    tree = etree.parse(filename)

    root = tree.getroot()
    if root.tag != "{http://www.w3.org/2000/svg}svg":
        raise TypeError("file %s does not seem to be a valid SVG file", filename)

    width = root.get("width")
    height = root.get("height")

    svg = transform_shape(parser(root))

    return svg


def transform_shape(element):
    # apply to current element
    element.apply_transform_matrix(element._transform_matrix)

    # apply to all its children
    for child in element.children:
        transform_shape(child)

    return element


def parser(element):
    shape = PShape([(0, 0)], children=[])

    transform = element.get("transform")
    transform_matrix = np.identity(4)
    if transform:
        properties = re.findall(r"(\w+)\(([\w\.]+)[,\s]*([^(]*)\)", transform)
        for p in properties:
            if p[0] == "scale":
                mat = matrix.scale_transform(float(p[1]), float(p[2]))
                transform_matrix = transform_matrix.dot(mat)
            elif p[0] == "translate":
                mat = matrix.translation_matrix(float(p[1]), float(p[2]))
                transform_matrix = transform_matrix.dot(mat)

    shape.transform_matrix(transform_matrix)

    for e in element:
        tag = e.tag.replace("{http://www.w3.org/2000/svg}", "")
        if tag in parser_function.keys():
            shape.add_child(parser_function[tag](e))
        elif tag == "g":
            shape.add_child(parser(e))
        else:
            continue

    return shape


def shape(shape, x=0, y=0):
    """
    Draws shapes to the display window

    :param shape: shape object
    :type shape: PShape

    :param x: x-coordinate of the shape
    :type x: float

    :param y: y-coordinate of the shape
    :type y: float
    """
    with transforms.push_matrix():
        transforms.translate(x, y)
        primitives.draw_shape(shape)



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from contextlib import AbstractContextManager
from copy import deepcopy

from . import p5

from contextlib import AbstractContextManager


class _StyleContext(AbstractContextManager):
    def __exit__(self, exc_type, exc_value, traceback):
        pop_style()


def push_style():
    """Save the current style settings and then restores them on exit.

    The 'style' information consists of all the parameters controlled
    by the following functions (the ones indicated by an asterisks '*'
    aren't available yet):

    - background
    - fill, no_fill
    - stroke, no_stroke
    - rect_mode
    - ellipse_mode
    - shape_mode
    - color_mode
    - tint
    - (*) stroke_weight
    - (*) stroke_cap
    - (*) stroke_join
    - (*) image_mode
    - (*) text_align
    - (*) text_font
    - (*) text_mode
    - (*) text_size
    - (*) text_leading
    -  emissive
    -  specular
    -  shininess
    -  ambient
    -  material

    """
    renderer_styles = deepcopy(p5.renderer.style)
    p5.renderer.style_stack.append(renderer_styles)
    return _StyleContext()


def pop_style():
    """Restores previously pushed style settings"""
    assert len(p5.renderer.style_stack) > 0, "No styles to pop"
    renderer_styles = p5.renderer.style_stack.pop()

    p5.renderer.style = renderer_styles



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Functions to work with Bezier curves and Catmul-Rom splines.

"""

from collections.abc import Iterable
from functools import wraps
from ..pmath import Point

__all__ = [
    # BEZIER METHODS
    "bezier_point",
    "bezier_tangent",
    "bezier_detail",
    # CURVE METHODS
    "curve_point",
    "curve_tangent",
    "curve_detail",
    "curve_tightness",
    # QUADRATIC METHODS
    "quadratic_point",
]

curve_resolution = 20

bezier_resolution = 20

curve_tightness_amount = 0

curve_basis_matrix = [
    [-0.5, 1.5, -1.5, 0.5],
    [1, -2.5, 2, -0.5],
    [-0.5, 0, 0.5, 0],
    [0, 1, 0, 0],
]


def typecast_arguments_as_points(func):
    """Typecast all but the last argument of the function as Points."""

    @wraps(func)
    def decorated(*args, **kwargs):
        new_args = [Point(*arg) for arg in args[:-1]]
        new_args.append(args[-1])
        ret_value = func(*new_args, **kwargs)
        return ret_value

    return decorated


def bezier_detail(detail_value):
    """Change the resolution used to draw bezier curves.

    :param detail_value: New resolution to be used.
    :type detail_value: int
    """
    global bezier_resolution
    bezier_resolution = max(1, detail_value)


def bezier_point(start, control_1, control_2, stop, parameter):
    """Return the coordinate of a point along a bezier curve.

    :param start: The start point of the bezier curve
    :type start: float or n-tuple

    :param control_1: The first control point of the bezier curve
    :type control_1: float or n-tuple

    :param control_2: The second control point of the bezier curve
    :type control_2: float or n-tuple

    :param stop: The end point of the bezier curve
    :type stop: float or n-tuple

    :param parameter: The parameter for the required location along
        the curve. Should be in the range [0.0, 1.0] where 0 indicates
        the start of the curve and 1 indicates the end of the curve.
    :type parameter: float

    :returns: The coordinate of the point along the bezier curve.
    :rtype: float or n-tuple

    """
    # Package floats into tuples
    is_iterable = isinstance(start, Iterable)
    if not is_iterable:
        start, control_1, control_2, stop = (
            (start,),
            (control_1,),
            (control_2,),
            (stop,),
        )

    t = parameter
    t_ = 1 - parameter
    P = [start, control_1, control_2, stop]
    coeffs = [t_ * t_ * t_, 3 * t * t_ * t_, 3 * t * t * t_, t * t * t]
    ans = tuple(sum(pt[i] * c for pt, c in zip(P, coeffs)) for i in range(len(start)))
    # Unpack answer if input is not iterable
    if not is_iterable:
        ans = ans[0]
    return ans


def bezier_tangent(start, control_1, control_2, stop, parameter):
    """Return the tangent at a point along a bezier curve.

    :param start: The start point of the bezier curve
    :type start: float or n-tuple.

    :param control_1: The first control point of the bezier curve
    :type control_1: float or n-tuple.

    :param control_2: The second control point of the bezier curve
    :type control_2: float or n-tuple.

    :param stop: The end point of the bezier curve
    :type stop: float or n-tuple.

    :param parameter: The parameter for the required tangent location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The tangent at the required point along the bezier
        curve.
    :rtype: float or n-tuple

    """
    # Package floats into tuples
    is_iterable = isinstance(start, Iterable)
    if not is_iterable:
        start, control_1, control_2, stop = (
            (start,),
            (control_1,),
            (control_2,),
            (stop,),
        )

    t = parameter

    def tangent(a, b, c, d):
        return (
            3 * t * t * (3 * b - 3 * c + d - a) + 6 * t * (a - 2 * b + c) + 3 * (b - a)
        )

    ans = tuple(
        tangent(start[i], control_1[i], control_2[i], stop[i])
        for i in range(len(start))
    )
    # Unpack answer if input is not iterable
    if not is_iterable:
        ans = ans[0]
    return ans


def _reinit_curve_matrices():
    # TODO: Add basis matrices for faster tessellation.
    pass


def curve_detail(detail_value):
    """Change the resolution used to draw bezier curves.

    :param detail_value: New resolution to be used.
    :type detail_value: int
    """
    global curve_resolution
    curve_resolution = detail_value


def curve_tightness(amount):
    """Change the curve tightness used to draw curves.

    :param amount: new curve tightness amount.
    :type amount: int
    """
    global curve_tightness_amount
    curve_tightness_amount = amount
    _reinit_curve_matrices()


def curve_point(point_1, point_2, point_3, point_4, parameter):
    """Return the coordinates of a point along a curve.

    :param point_1: The first control point of the curve.
    :type point_1: float or n-tuple.

    :param point_2: The second control point of the curve.
    :type point_2: float or n-tuple.

    :param point_3: The third control point of the curve.
    :type point_3: float or n-tuple.

    :param point_4: The fourth control point of the curve.
    :type point_4: float or n-tuple.

    :param parameter: The parameter for the required point location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The coordinate of the point at the required location
        along the curve.
    :rtype: float or n-tuple

    """
    # Package floats into tuples
    is_iterable = isinstance(point_1, Iterable)
    if not is_iterable:
        point_1, point_2, point_3, point_4 = (
            (point_1,),
            (point_2,),
            (point_3,),
            (point_4,),
        )

    t = parameter
    basis = curve_basis_matrix
    P = [point_1, point_2, point_3, point_4]
    coeffs = [sum(t ** (3 - i) * basis[i][j] for i in range(4)) for j in range(4)]
    ans = tuple(sum(pt[i] * c for pt, c in zip(P, coeffs)) for i in range(len(point_1)))
    # Unpack answer if input is not iterable
    if not is_iterable:
        ans = ans[0]
    return ans


def curve_tangent(point_1, point_2, point_3, point_4, parameter):
    """Return the tangent at a point along a curve.

    :param point_1: The first control point of the curve.
    :type point_1: float or n-tuple.

    :param point_2: The second control point of the curve.
    :type point_2: float or n-tuple.

    :param point_3: The third control point of the curve.
    :type point_3: float or n-tuple.

    :param point_4: The fourth control point of the curve.
    :type point_4: float or n-tuple.

    :param parameter: The parameter for the required tangent location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The tangent at the required point along the curve.
    :rtype: float or n-tuple

    """
    # Package floats into tuples
    is_iterable = isinstance(point_1, Iterable)
    if not is_iterable:
        point_1, point_2, point_3, point_4 = (
            (point_1,),
            (point_2,),
            (point_3,),
            (point_4,),
        )

    t = parameter
    basis = curve_basis_matrix
    P = [point_1, point_2, point_3, point_4]
    coeffs = [
        sum((3 - i) * (t ** (2 - i)) * basis[i][j] for i in range(3)) for j in range(4)
    ]
    ans = tuple(sum(pt[i] * c for pt, c in zip(P, coeffs)) for i in range(len(point_1)))
    # Unpack answer if input is not iterable
    if not is_iterable:
        ans = ans[0]
    return ans


def quadratic_point(start, control, stop, parameter):
    """Return the coordinates of a point along a bezier curve.

    :param point_1: The start point of the curve.
    :type point_1: float or n-tuple.

    :param point_3: The control point of the curve.
    :type point_3: float or n-tuple.

    :param point_4: The end point of the curve.
    :type point_4: float or n-tuple.

    :param parameter: The parameter for the required point location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The coordinate of the point at the required location
        along the curve.
    :rtype: float or n-tuple

    """
    is_iterable = isinstance(start, Iterable)
    if not is_iterable:
        start, control, stop = (start,), (control,), (stop,)

    t = parameter
    t_ = 1 - parameter
    P = [start, control, stop]
    coeffs = [t_ * t_, 2 * t * t_, t * t]
    ans = tuple(sum(pt[i] * c for pt, c in zip(P, coeffs)) for i in range(len(start)))
    # Unpack answer if input is not iterable
    if not is_iterable:
        ans = ans[0]
    return ans


# Set the default values.
bezier_detail(20)
curve_detail(20)
curve_tightness(0)



#!/usr/bin/env python

import os
import sys

from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))

packages = ["p5"]

with open("requirements.txt", "r", encoding="UTF-8") as f:
    requires = f.read().split("\n")

meta_data = {}

with open(os.path.join(here, "p5", "__version__.py"), "r", encoding="utf-8") as f:
    exec(f.read(), meta_data)

with open("README.rst", "r", encoding="utf-8") as f:
    readme = f.read()

setup(
    name=meta_data["__title__"],
    version=meta_data["__version__"],
    description=meta_data["__description__"],
    long_description=readme,
    long_description_content_type="text/x-rst",
    author=meta_data["__author__"],
    url=meta_data["__url__"],
    author_email=meta_data["__author_email__"],
    license=meta_data["__license__"],
    packages=find_packages(),
    package_data={
        "": ["LICENSE"],
    },
    package_dir={"p5": "p5"},
    include_package_data=True,
    install_requires=requires,
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Topic :: Artistic Software",
        "Topic :: Education",
        "Topic :: Multimedia",
        "Topic :: Multimedia :: Graphics",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



"""
This file contains the handlers needed for glfw event handling
"""


import builtins
import glfw
from p5.core import p5
from p5.sketch.events import KeyEvent, MouseEvent

from dataclasses import dataclass
from time import time

from p5.util import keys

BUTTONMAP = {
    glfw.MOUSE_BUTTON_LEFT: 1,
    glfw.MOUSE_BUTTON_RIGHT: 2,
    glfw.MOUSE_BUTTON_MIDDLE: 3,
}


KEYMAP = {
    glfw.KEY_LEFT_SHIFT: keys.SHIFT,
    glfw.KEY_RIGHT_SHIFT: keys.SHIFT,
    glfw.KEY_LEFT_CONTROL: keys.CONTROL,
    glfw.KEY_RIGHT_CONTROL: keys.CONTROL,
    glfw.KEY_LEFT_ALT: keys.ALT,
    glfw.KEY_RIGHT_ALT: keys.ALT,
    glfw.KEY_LEFT_SUPER: keys.META,
    glfw.KEY_RIGHT_SUPER: keys.META,
    glfw.KEY_LEFT: keys.LEFT,
    glfw.KEY_UP: keys.UP,
    glfw.KEY_RIGHT: keys.RIGHT,
    glfw.KEY_DOWN: keys.DOWN,
    glfw.KEY_PAGE_UP: keys.PAGEUP,
    glfw.KEY_PAGE_DOWN: keys.PAGEDOWN,
    glfw.KEY_INSERT: keys.INSERT,
    glfw.KEY_DELETE: keys.DELETE,
    glfw.KEY_HOME: keys.HOME,
    glfw.KEY_END: keys.END,
    glfw.KEY_ESCAPE: keys.ESCAPE,
    glfw.KEY_BACKSPACE: keys.BACKSPACE,
    glfw.KEY_F1: keys.F1,
    glfw.KEY_F2: keys.F2,
    glfw.KEY_F3: keys.F3,
    glfw.KEY_F4: keys.F4,
    glfw.KEY_F5: keys.F5,
    glfw.KEY_F6: keys.F6,
    glfw.KEY_F7: keys.F7,
    glfw.KEY_F8: keys.F8,
    glfw.KEY_F9: keys.F9,
    glfw.KEY_F10: keys.F10,
    glfw.KEY_F11: keys.F11,
    glfw.KEY_F12: keys.F12,
    glfw.KEY_SPACE: keys.SPACE,
    glfw.KEY_ENTER: keys.ENTER,
    "\r": keys.ENTER,
    glfw.KEY_TAB: keys.TAB,
}

MOD_KEYS = [keys.SHIFT, keys.ALT, keys.CONTROL, keys.META]


@dataclass
class InputState:
    """
    A class to store the current state of Input
    """

    modifiers: list
    last_mouse_clicked_time: float
    next_key_events: list
    next_key_text: dict

    def process_mod(self, key, down):
        """Process (possible) keyboard modifiers"""
        if key in MOD_KEYS:
            if down:
                if key not in self.modifiers:
                    self.modifiers.append(key)
            elif key in self.modifiers:
                self.modifiers.pop(self.modifiers.index(key))
        return self.modifiers

    def process_key(self, key):
        if 32 <= key <= 127:
            return keys.Key(chr(key)), chr(key)
        elif key in KEYMAP:
            return KEYMAP[key], ""
        else:
            return None, ""


class PseudoKeyEvent:
    """Class to pass pseudo event data. They are used to work with the events module which is vispy dependent"""

    def __init__(self, key=None, mod=None, text=None):
        self.key = key
        self.modifiers = mod
        self.text = text


class PseudoMouseEvent:
    """
    Class to pass pseudo event data. They are used to work with the events module which is vispy dependent
    """

    def __init__(
        self, pos=(0, 0), delta=(0, 0), modifiers=None, button=None, buttons=None
    ):
        """Initialize the pseduoMouseEventClass to work with exisiting event system

        Args:
            pos (tuple): Position of the cursor
            delta (tuple): The increment by which the mouse wheel has moved
            modifiers (list): Modifiers active during the event
            button (int): Mouse Button that was pressed
            buttons (list): Mouse Buttons that were pressed during the event
        """
        self.pos = pos
        self.delta = delta
        self.modifiers = modifiers
        self.button = button
        self.buttons = buttons


# Hold the input state for current sketch
input_state = InputState([], 0, [], {})


def on_mouse_button(window, button, action, mod):
    pos = glfw.get_cursor_pos(window)

    if button < 3:
        button = BUTTONMAP.get(button, 0)

    event = PseudoMouseEvent(pos, (0, 0), input_state.modifiers, button, [])
    mev = MouseEvent(event, active=(action == glfw.PRESS))

    if action == glfw.PRESS:
        p5.sketch._enqueue_event("mouse_pressed", mev)
    else:
        p5.sketch._enqueue_event("mouse_released", mev)
        p5.sketch._enqueue_event("mouse_clicked", mev)

        dt_max = 0.3  # time in seconds for a double-click detection
        if time() - input_state.last_mouse_clicked_time <= dt_max:
            p5.sketch._enqueue_event("mouse_double_clicked", mev)

        input_state.last_mouse_clicked_time = time()


def on_mouse_scroll(window, x_off, y_off):
    pos = glfw.get_cursor_pos(window)
    delta = (float(x_off), float(y_off))
    event = PseudoMouseEvent(pos=pos, delta=delta, modifiers=input_state.modifiers)
    mev = MouseEvent(event, active=builtins.mouse_is_pressed)

    p5.sketch._enqueue_event("mouse_wheel", mev)


def on_mouse_motion(window, x, y):
    event = PseudoMouseEvent(pos=(x, y), modifiers=input_state.modifiers)
    mev = MouseEvent(event, active=builtins.mouse_is_pressed)

    # Queue a 'mouse_dragged` or `mouse_moved` event, not both similar to p5.js
    p5.sketch._enqueue_event(
        "mouse_dragged" if builtins.mouse_is_pressed else "mouse_moved", mev
    )


def on_key_press(window, key, scancode, action, mod):
    key, text = input_state.process_key(key)
    if action == glfw.PRESS:
        down = True
    elif action == glfw.RELEASE:
        down = False
    else:
        return
    input_state.process_mod(key, down=down)

    # NOTE: GLFW only provides localized characters via _on_key_char, so if
    # this event contains a character we store all other data and dispatch
    # it once the final unicode character is sent shortly after.
    if text != "" and action == glfw.PRESS:
        input_state.next_key_events.append((action, key, input_state.modifiers))
    else:
        if key in input_state.next_key_text:
            text = input_state.next_key_text[key]
            del input_state.next_key_text[key]
        # Not a char event, call the event normally
        event = PseudoKeyEvent(key, input_state.modifiers, text)
        kev = KeyEvent(event, action == glfw.PRESS)
        p5.sketch._enqueue_event(
            "key_pressed" if action == glfw.PRESS else "key_released", kev
        )


def on_key_char(window, text):
    # Repeat strokes (frequency configured at OS) are sent here only,
    # no regular _on_key_press events. Currently ignored!
    if len(input_state.next_key_events) == 0:
        return

    (action, key, mod) = input_state.next_key_events.pop(0)
    input_state.next_key_text[key] = text

    event = PseudoKeyEvent(key, mod, chr(text))
    kev = KeyEvent(event, action == glfw.PRESS)

    p5.sketch._enqueue_event(
        "key_pressed" if action == glfw.PRESS else "key_released", kev
    )


def on_window_focus(window, focused):
    builtins.focused = bool(focused)


def on_close(window):
    pass



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""
3D geometry class for p5py
"""

from . import p5
import numpy as np
import math


class Geometry:
    """
        Geometry class for all 3D shapes

    :param detail_x: number of triangle subdivisions in x-dimension
    :type detail_x: integer

    :param detail_y: number of triangle subdivisions in y-dimension
    :type detail_y: integer

    """

    def __init__(self, detail_x=1, detail_y=1):
        self.vertices = []

        self.line_vertices = []

        self.line_normals = []

        self.vertex_normals = []

        self.faces = []

        self.uvs = []
        # a 2D array containing edge connectivity pattern for create line vertices
        # based on faces for most objects
        self.edges = []
        self.detail_x = detail_x
        self.detail_y = detail_y

        self.stroke_indices = []

        self.matrix = np.identity(4)
        self.material = p5.renderer.style.material

    def reset(self):
        """
        Reset geometry parameters
        """
        self.vertices = []
        self.line_vertices = []
        self.line_normals = []
        self.vertex_normals = []
        self.faces = []
        self.uvs = []
        self.edges = []

    def compute_faces(self):
        """
        Adds the faces for the geometry for predefined order of vertices
        """
        self.faces = []
        sliceCount = self.detail_x + 1
        for i in range(self.detail_y):
            for j in range(self.detail_x):
                a = i * sliceCount + j
                b = i * sliceCount + j + 1
                c = (i + 1) * sliceCount + j + 1
                d = (i + 1) * sliceCount + j
                self.faces.append([a, b, d])
                self.faces.append([d, b, c])

    def make_triangle_edges(self):
        """
        Adds the edges to the geometry based on the faces
        """
        self.edges = []
        for j in range(len(self.faces)):
            self.edges.append([self.faces[j][0], self.faces[j][1]])
            self.edges.append([self.faces[j][1], self.faces[j][2]])
            self.edges.append([self.faces[j][2], self.faces[j][0]])

    def get_face_normal(self, faceId):
        """
        Returns the normal for a given face
        """
        face = self.faces[faceId]
        vA = np.array(self.vertices[face[0]])
        vB = np.array(self.vertices[face[1]])
        vC = np.array(self.vertices[face[2]])
        ab = vB - vA
        ac = vC - vA
        n = np.cross(ab, ac)
        ln = np.linalg.norm(n)

        sinAlpha = ln / (np.linalg.norm(ab) * np.linalg.norm(ac))
        if sinAlpha > 1:
            sinAlpha = 1

        return n * math.sin(sinAlpha) / ln

    def compute_normals(self):
        """
        Compute normals for every vertex
        """
        self.vertex_normals = []
        for iv in range(len(self.vertices)):
            self.vertex_normals.append([0, 0, 0])

        for f in range(len(self.faces)):
            face = self.faces[f]
            face_normal = self.get_face_normal(f)

            for fv in range(3):
                vertex_index = face[fv]
                self.vertex_normals[vertex_index] += face_normal

        for iv in range(len(self.vertices)):
            self.vertex_normals[iv] = self.vertex_normals[iv] / np.linalg.norm(
                self.vertex_normals[iv]
            )

    def edges_to_vertices(self):
        self.line_vertices = []
        self.line_normals = []

        for i in range(len(self.edges)):
            begin = self.vertices[self.edges[i][0]]
            end = self.vertices[self.edges[i][1]]

            direction = np.array(end) - np.array(begin)
            direction = direction / np.linalg.norm(direction)
            direction = direction.tolist()

            a = begin
            b = begin
            c = end
            d = end
            dirAdd = direction
            dirSub = direction
            dirAdd.append(1)
            dirSub.append(-1)

            self.line_normals.extend([dirAdd, dirSub, dirAdd, dirAdd, dirSub, dirSub])
            self.line_vertices.extend([a, b, c, c, b, d])



# This file adds Processing API compatibility for p5py

from .http_data import http_post, http_get
from .local_data import load_table


def httpGet(url):
    """
    Performs a synchronous GET request on the given url.

    :param url: url of the API
    :type url: string

    """
    return http_get(url)


def httpPost(url, postData):
    """
    Performs a synchronous POST request on the given url.

    :param url: url of the API
    :type url: string
    :param postData: Key-value pairs of the request payload.
    :type postData: dict

    """
    return http_post(url, postData)


def loadTable(path, mode="csv"):
    """
    Calls Table class and returns a Table class object

    :param path: Path to file
    :type path: string

    :param mode: Type of File csv/ssv/tsv.
    :type mode: string

    :returns: A table object
    :rtype: object

    """
    load_table(path, mode)



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Shaders used by the main program"""

from collections import namedtuple
from ..util import read_shader

ShaderSource = namedtuple("ShaderSource", "vert frag")

src_default = ShaderSource(
    read_shader("Vispy2DRenderer/shaders/2d/default2d.vert"),
    read_shader("Vispy2DRenderer/shaders/common/default.frag"),
)
src_texture = ShaderSource(
    read_shader("Vispy2DRenderer/shaders/2d/texture.vert"),
    read_shader("Vispy2DRenderer/shaders/2d/texture.frag"),
)
# Shader sources to draw framebuffers textues.
src_fbuffer = ShaderSource(
    read_shader("Vispy2DRenderer/shaders/common/fbuffer.vert"),
    read_shader("Vispy2DRenderer/shaders/common/fbuffer.frag"),
)
src_line = ShaderSource(
    read_shader("Vispy2DRenderer/shaders/2d/stroke.vert"),
    read_shader("Vispy2DRenderer/shaders/2d/stroke.frag"),
)



import pkgutil
import os
from ..core import p5
import builtins


def read_shader(filename):
    """Reads a shader in string mode and returns the content"""

    return pkgutil.get_data("p5", os.path.join("sketch/", filename)).decode()


def ensure_p3d(name):
    assert p5.mode == "P3D", name + "is only available in P3D renderer"


def scale_tuple(t, scale=255):
    """Divides each element of tuple by scale"""
    return tuple(x / scale for x in t)
